/*
 * Copyright (c) 2015, Majenko Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * * Neither the name of Majenko Technologies nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

 /* well, the above wrote the wifi stack, I wrote the rest and it's GPL-V2, C Doug Coulter 2016 */

 // test change 12/15 retest - very strange, think I found a bug.  Remebers path to "save as" last time, but when trying to get it back..not correct.

 //@@@ for later use:  ESP.restart() is software reset, but doesn't work on serial jig - only later on.

 // rewriting this code to support DDS instead of relays and sht31 5/18/17

#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <WiFiUdp.h>
#include <AD9850.h> // for ad9850 dds module

// for use in dtostrf
#define NUM_SIZE 15
#define D_WIDTH (-(NUM_SIZE-1))

// pin defines as this might get ported to another platform
//#define SDA 4
//#define SCL 5
#define DEBUG_PIN 16

// AD9850 Module....
// set pin numbers:
#define W_CLK 12
#define FQ_UD 13 
#define DATA 14
#define RESET 2

double current_freq; // the current frequency
double startf; // start frequency
double endf; // end frequency
double incr; // multiplier for frequency
unsigned long update_rate; // ms
bool do_update;
bool DDS_Enable;
bool DDS_Enable_z; // for quickly detecting and acting on the "off" command


// scheduling stuff
bool debug; // if pin is high, we're not running on the dev box
bool did_something; // we did something this loop, so hold off till next one

bool do_tellem;
//bool update_th;
//bool update_plot_data;

unsigned long loop_millis;
#define MINUTE_MS 60000
unsigned long minute_z;
unsigned long DDS_z;

// tellem stuff
WiFiUDP tport;
char packetBuffer[255];
#define TELLEM_PORT 53831
IPAddress myip; // my ip as assigned by dhcp
IPAddress bcip; // broadcast addr for this network
unsigned char mac[6]; // my mac, I hope
char tellem_string[40]; // What we'll broadcast
unsigned long tellems;
unsigned long errors;


// WAP stuff - uh-oh hardcoded creds @@@
const char *ssid = "machswap"; //machswap
const char *password = "fusordoug";

ESP8266WebServer server ( 80 );

////////////////////////////////////
void init_ram()
{ // make ram clean
  word i;
  debug = did_something = do_tellem = do_update = DDS_Enable = DDS_Enable_z = false;
  tellems = errors = 0;
  minute_z = DDS_z =  millis();
  current_freq = 0; // the current frequency (default)
  startf = 5000; // start frequency
  endf = 2500000; // end frequency
  incr = 1.1; // multiplier for frequency
  update_rate = 100; // ms

}

////////////////////////////////////
void build_tellem_string ()
{
  char i;
  String mystring;

//@@@ class String can't 0 pad hex...
// but sprintf and strncpy are lots more lines of code (less rom though)
  mystring = myip.toString();
  mystring += " ESP_";
  for (i=3;i<6;i++) 
  {
    if (16 > mac[i]) mystring += "0"; // yes, it is that dumb    
    mystring += String(mac[i],HEX);
   }
  mystring.toUpperCase(); // because damn, that's what it is on the net for some retarded opsys
  mystring.toCharArray(tellem_string, 39);
  
  Serial.println(tellem_string); // debug
}
////////////////////////////////////
void tellem()
{
  tport.beginPacket(bcip,TELLEM_PORT);
  tport.write(tellem_string);
  tport.endPacket(); 
  tellems++; // for grins, count them
  do_tellem = false; // we did it
}
////////////////////////////////////
void ditch_incoming()
{ // ignore other broadcasts, maybe stop possible memory leak.
  int packetSize = tport.parsePacket();
  if (packetSize) {
    int len = tport.read(packetBuffer, 254);
    if (debug)
    { // but hey, make sure it works if we are in debug mode!
     if (len > 0) packetBuffer[len] = 0;
     Serial.println(packetBuffer);
    }  
  }
}
////////////////////////////////////
void handleRoot() {
  #define PAGE_SIZE 600
	char temp[PAGE_SIZE]; // 500 was too short!
  char c_freq[NUM_SIZE];
  char c_startf[NUM_SIZE];
  char c_endf[NUM_SIZE];
  char c_incr[NUM_SIZE];

  dtostrf(current_freq,D_WIDTH,3,c_freq);
  dtostrf(startf,D_WIDTH,3,c_startf);
  dtostrf(endf,D_WIDTH,3,c_endf);
  dtostrf(incr,D_WIDTH,3,c_incr);
   
  did_something = true; // we did something that takes time
	snprintf ( temp, PAGE_SIZE,
"<html>\
  <head>\
    <meta http-equiv='refresh' content='60'/>\
    <title>ESP DDS</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
  <body>\
    <h1>Digitally controled frequency generator</h1>\
    <p>Bcasts:%d Sensor Errors:%d\tWiFi Signal:%d dBm</p>\
    <p>Current Frequency: %s</p>\
    <p>Start Frequency: %s</p>\
    <p>End Frequency: %s</p>\
    <p>Increment: %s</p>\
    <p>Update mS: %d</p>\
    <p>DDS Enable: %d</p>\
    <p><a href='/controls'>Controls</a></p>\
  </body>\
</html>"
 ,tellems,errors,WiFi.RSSI(),c_freq,c_startf,c_endf,c_incr,update_rate,DDS_Enable);
	server.send ( 200, "text/html", temp );
}
/////////////////////////////////////////////////
void handleControls()
{
 int i,j;
 String page;

  if (j = server.args()) // will only be true for post-submit, not refresh 
  { //     due to hidden field will always be true in that case
   for (i=0;i<j;i++)
   { // compare names to stuff, fix up input errors, set vars in memory    
    if (server.argName(i) == "startf")
    {
      startf = server.arg(i).toFloat();
      startf = abs(startf);
      if (startf < 0.3) startf = 0.3; // ok, we could make this the actual limit, but really?
    }
    if (server.argName(i) == "endf")
    {
      endf = server.arg(i).toFloat();
      endf = abs(endf);
      if (endf < 0.3) endf = 0.3; // ok, we could make this the actual limit, but really?
    }
    if (server.argName(i) == "incr")
    {
      incr = server.arg(i).toFloat();
      incr = abs(incr);
    } // check later to see if right direction increment...
    if (server.argName(i) == "update_rate")
    {
//      Serial.print("at update_rate:");
//      Serial.println(update_rate);
      update_rate = server.arg(i).toInt();
      update_rate = abs(update_rate);
      if (update_rate <= 1 ) update_rate = 111; // pick something not utterly stupid
    }
    if (server.argName(i) == "enable") DDS_Enable = true;
    else DDS_Enable = false;
    
   }// for all variables
   // some fixups for operator oopsies
   if (startf < endf) // normal case?
   {
    if (incr <=1) incr = 1/incr;
   }
   if (endf < startf)
   {
    if (incr >=1) incr = 1/incr;
   }
   // I'm guessing other logic in loop will handle the case of incr=1.0 or startf=endf fine
  }

 page = \
 "<html>\
  <head>\
    <title>DDS Ctl</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
  <body>\
    <h1>DDS Controls</h1><p> </p>\
    <FORM action=\"/controls\" method=\"post\"> <INPUT type=\"hidden\" name=\"user\" value=\"1\"> ";
// create some input controls
 page += "Sweep start frequency: <INPUT type=\"text\" size=\"NUM_SIZE\" maxlength=\"NUM_SIZE\" name=\"startf\" value=\"";
 page += String(startf,3);
 page += "\"> Hz <br>";

 page += "Sweep end frequency:  <INPUT type=\"text\" size=\"NUM_SIZE\" maxlength=\"NUM_SIZE\" name=\"endf\" value=\"";
 page += String(endf,3);
 page += "\"> Hz <br>";

 page += "Sweep update multiplier: <INPUT type=\"text\" size=\"NUM_SIZE\" maxlength=\"NUM_SIZE\" name=\"incr\" value=\"";
 page += String(incr,3);
 page += "\"> percent/100 <br>";

 page += "Sweep update time: <INPUT type=\"text\" size=\"NUM_SIZE\" maxlength=\"NUM_SIZE\" name=\"update_rate\" value=\"";
 page += String(update_rate,DEC);
 page += "\"> milliseconds <br>";

 if(DDS_Enable) page += "<INPUT type=\"checkbox\" name=\"enable\" value=\"1\" checked>DDS Enable<br>";
 else page += "<INPUT type=\"checkbox\" name=\"enable\" value=\"1\" >DDS Enable<br>";

//then

  page += \  
"<p></p><INPUT type=\"submit\" value=\"Do it\"></FORM>\ 
    <p> </p><a href='/'>Home</a></p>\
  </body>\
</html>";
 
  did_something = true; // we did something that takes time
  server.send ( 200, "text/html", page );    
}
////////////////////////////////////////////

void handleNotFound() {
	String message = "File Not Found\n\n";
	message += "URI: ";
	message += server.uri();
	message += "\nMethod: ";
	message += ( server.method() == HTTP_GET ) ? "GET" : "POST";
	message += "\nArguments: ";
	message += server.args();
	message += "\n";

	for ( uint8_t i = 0; i < server.args(); i++ ) {
		message += " " + server.argName ( i ) + ": " + server.arg ( i ) + "\n";
	}

	server.send ( 404, "text/plain", message );
  did_something = true;
}
// *****************************************************************************/
void setup ( void ) {
//  Wire.pins(SDA,SCL);
   init_ram();
// decide on debug mode here
   pinMode (DEBUG_PIN,INPUT);
   debug = !digitalRead(DEBUG_PIN);

	 Serial.begin ( 115200 );
	 WiFi.begin ( ssid, password );
	 Serial.println ( "" );

	// Wait for connection
	 while ( WiFi.status() != WL_CONNECTED ) {
		delay ( 500 );
		Serial.print ( "." );
   /// tellem stuff
    WiFi.macAddress(mac); // get my MAC addr, as it's part of hostname
    myip = WiFi.localIP(); // class and union of dword and 4 bytes
    bcip = myip; // copy net address but
    bcip[3] = 255; // make it broadcast address
	}

	Serial.println ( "" );
	Serial.print ( "Connected to " );
	Serial.println ( ssid );
	Serial.print ( "My IP address: " );
	Serial.println ( WiFi.localIP() );


  server.on ( "/", handleRoot );
	server.on ( "/inline", []() {
	server.send ( 200, "text/plain", "this is an operator goof" );
	} );
  server.on ( "/controls",handleControls);
	server.onNotFound ( handleNotFound );
	server.begin();
	Serial.println ( "HTTP server started" );
  // tellem stuff
  build_tellem_string();
  tport.begin(TELLEM_PORT); // for tellem
  DDS.begin(W_CLK,FQ_UD,DATA,RESET); // set up ad9850 module wiring
}
///////////////////////////////////////////////////////////////////
void loop ( void ) {
  did_something = false; // because we haven't yet on this round
	server.handleClient(); // if any hits should set did_something to true
  loop_millis = millis(); // just do this once
  debug = !digitalRead(DEBUG_PIN); // semi-realtime update of debug

// set scheduling flags for various intervals
// add setting various bools to these to get something to run (eventually)
  if (loop_millis - minute_z >= MINUTE_MS)
  {
    minute_z = loop_millis; // update for next time
    do_tellem = true; // schedule 1/minute stuff
  }
  if (loop_millis - DDS_z >= update_rate) // dds sweep update
  {
   DDS_z = loop_millis;
   do_update = true;   
  }

  if (!DDS_Enable && DDS_Enable_z) // on but got turned off
  { // so do that immediately
//@@@ do that here

    current_freq = 0; // we're stopping running, right?
    DDS_Enable_z = DDS_Enable; // we got it
    did_something = true; // worry about anything else next round
  } else if (DDS_Enable && !DDS_Enable_z) // off but got turned on
  {
   current_freq = startf; // set to begin sweep
   did_something = true;
   DDS_Enable_z = DDS_Enable; // we got it
  }

  
// Do stuff.  Not 1::1 with how often requested on purpose - important stuff gets priority regardless.
// (because we check/do that first)  
// If we've already "done something" this loop, wait for next roundy round.

  if (do_update && !did_something)
  { 
   DDS.setfreq(current_freq,0); // phase always zero for now
   current_freq *= incr; // get ready for next output
   // test to see if out of bounds
   if (incr > 1.0) // sweeping up so check if greater than endf
   {
    if (current_freq > endf) current_freq = startf; // start over
   } else // sweeping down
   {
    if (current_freq < endf) current_freq = startf;
   }
   do_update = false; // did it already
   did_something = true; 
  }

  if (do_tellem  && !did_something) // only if it is time AND we didn't do something else first 
  {
   tellem();
   did_something = true; // skip other things this loop
  } // every minute or so

 
   if (!did_something) ditch_incoming(); // in case it matters?
// fancy opsys, eh?
}
//********************************************************************************************

